from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING

from . import (
    TEXT,
    RunException,
    fail,
    info,
    run,
)


if TYPE_CHECKING:
    import argparse


def django_available() -> bool:
    try:
        run("python", "-m", "django", "--version")
    except RunException:
        return False
    return True


def get_django_apps() -> set[str]:
    try:
        return set(
            run(
                "python",
                "manage.py",
                "shell",
                "-c",
                "from django.apps import apps; "
                "print(','.join(list(apps.app_configs.keys())))",
            )
            .strip()
            .split(",")
        )
    except RunException as ex:
        fail(str(ex))


def get_migrations(branch: str) -> defaultdict[str, list[str]]:
    try:
        files = sorted(run("git", "ls-tree", "-r", "--name-only", branch).split())
    except RunException as ex:
        fail(str(ex))

    app_to_migrations = defaultdict(list)
    for path in files:
        if "/migrations/" not in path:
            continue
        app, migration = path.split("/migrations/")
        app = app.split("/")[-1]
        if migration == "__init__.py":
            continue
        app_to_migrations[app].append(migration)

    return app_to_migrations


def find_youngest_shared_migration(list1: list[str], list2: list[str]) -> str:
    youngest = "zero"
    for migration1, migration2 in zip(list1, list2):
        if migration1 != migration2:
            break
        youngest = migration1.split("_", 1)[0]
    return youngest


def reverse_migrations(args: argparse.Namespace) -> None:
    if not django_available():
        fail("Django not found, won't reverse any migrations 🤷")

    try:
        current_branch = run("git", "rev-parse", "--abbrev-ref", "HEAD").strip()
    except RunException as ex:
        fail(str(ex))

    if current_branch == args.dst_branch:
        # no need to reverse if the current branch is the destination branch
        return

    info("Getting a list of Django applications...")
    available_apps = get_django_apps()

    all_current_migrations = get_migrations(current_branch)
    all_dest_migrations = get_migrations(args.dst_branch)

    info(
        f"👀 Looking for migrations to reverse on "
        f"{TEXT.UNDERLINE}{current_branch}{TEXT.END}...",
        delete_last_line=True,
    )
    migrations_reversed = 0
    for app, current_migrations in all_current_migrations.items():
        if app not in available_apps:
            # additional check against migration files for an application not
            # registered in the project.
            continue

        dest_migrations = all_dest_migrations[app]
        youngest = find_youngest_shared_migration(current_migrations, dest_migrations)

        if current_migrations[-1].startswith(youngest):
            # for a given application the last migration is the same on both branches,
            # there is no need to perform expensive reversals
            continue

        info(
            f"⚠️ reversing migrations for {TEXT.BOLD}{app}{TEXT.END} (up to {youngest})"
        )
        try:
            run("python", "manage.py", "migrate", app, youngest)
        except RunException as ex:
            fail(str(ex))

        migrations_reversed += 1

    if migrations_reversed:
        info(
            f"I have reversed migrations for {migrations_reversed} "
            f"app{'s' if migrations_reversed > 1 else ''}!"
        )

    else:
        info("Great Scott! No migrations to reverse!", delete_last_line=True)



from __future__ import annotations

import subprocess
import sys
from typing import NoReturn


__version__ = "0.1.0"


class RunException(Exception):
    pass


class TEXT:
    PURPLE = "\033[95m"
    CYAN = "\033[96m"
    DARKCYAN = "\033[36m"
    BLUE = "\033[94m"
    GREEN = "\033[92m"
    YELLOW = "\033[93m"
    RED = "\033[91m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"
    END = "\033[0m"


def run(command: str, *args: str) -> str | None:
    sp = subprocess.run([command, *args], capture_output=True, text=True)
    if sp.returncode != 0:
        raise RunException(sp.stderr)
    return sp.stdout


def fail(msg: str, status: int = 1) -> NoReturn:
    print(f"☠️ {msg}")
    exit(status)


def info(msg: str, delete_last_line: bool = False) -> None:
    if delete_last_line:
        sys.stdout.write("\x1b[1A\x1b[2K")
    print(f"{msg}")



from __future__ import annotations

import sys
from pathlib import Path
from typing import TYPE_CHECKING

from . import (
    RunException,
    fail,
    info,
    run,
)


if TYPE_CHECKING:
    import argparse


gs_pre_checkout_template = """#!/usr/bin/env bash
# this checks if HEAD is detached, 0 means normal 1 is detached
if git symbolic-ref -q HEAD > /dev/null;
then
    export GS_PRE_CHECKOUT=true
    export GS_DST_BRANCH=`git rev-parse --abbrev-ref HEAD`
    git switch --quiet -
    {script_path} reverse --dst-branch "$GS_DST_BRANCH"
    git switch --quiet -
fi
"""


post_checkout_template = """#!/usr/bin/env bash
{invoke_pre_checkout}
"""

invoke_pre_checkout = (
    'if [ -n "$GS_PRE_CHECKOUT" ] && "$GS_PRE_CHECKOUT"; '
    "then exit 0; "
    "else . .git/hooks/gs-pre-checkout; fi"
)

python_path = sys.executable
script_path = Path(python_path).with_name("great-scott")


def get_git_hooks_path() -> Path:
    try:
        return (
            Path(run("git", "rev-parse", "--show-toplevel").strip()) / ".git" / "hooks"
        )
    except RunException:
        fail(
            "fatal: not a git repository (or any of the parent directories), "
            "quitting..."
        )


def _install_post_checkout(git_hooks_path: Path) -> None:
    """Install post-checkout script"""
    post_checkout_path = git_hooks_path / "post-checkout"
    if not post_checkout_path.exists():
        with open(post_checkout_path, "w") as f:
            f.write(
                post_checkout_template.format(invoke_pre_checkout=invoke_pre_checkout)
            )
        post_checkout_path.chmod(0o755)
        return

    with open(post_checkout_path) as f:
        post_checkout_script_lines = list(map(lambda s: s.strip(), f.readlines()))

    if (
        not post_checkout_script_lines[0].startswith("#!")
        or "/bin/sh" not in post_checkout_script_lines[0]
        and "bash" not in post_checkout_script_lines[0]
    ):
        fail("post-checkout is not a bash or a sh script, quitting...")

    if invoke_pre_checkout in post_checkout_script_lines:
        # call is already in the script
        return

    post_checkout_script_lines.insert(1, invoke_pre_checkout)
    with open(post_checkout_path, "w") as f:
        f.write("\n".join(post_checkout_script_lines) + "\n")


def _uninstall_post_checkout(git_hooks_path: Path) -> None:
    """Remove call to pre-checkout from post-checkout script"""
    post_checkout_path = git_hooks_path / "post-checkout"
    if not post_checkout_path.exists():
        return

    with open(post_checkout_path) as f:
        post_checkout_script_lines = list(map(lambda s: s.strip(), f.readlines()))

    try:
        post_checkout_script_lines.remove(invoke_pre_checkout)
    except ValueError:
        return

    with open(post_checkout_path, "w") as f:
        f.write("\n".join(post_checkout_script_lines) + "\n")


def _install_pre_checkout(git_hooks_path: Path) -> None:
    """Install pre-checkout script"""
    pre_checkout_path = git_hooks_path / "gs-pre-checkout"
    with open(pre_checkout_path, "w") as f:
        f.write(
            gs_pre_checkout_template.format(
                script_path=script_path,
            )
        )
    pre_checkout_path.chmod(0o755)


def _uninstall_pre_checkout(git_hooks_path: Path) -> None:
    """Remove gs-pre-checkout script"""
    pre_checkout_path = git_hooks_path / "gs-pre-checkout"
    if pre_checkout_path.exists():
        pre_checkout_path.unlink()


def install(args: argparse.Namespace) -> None:
    info("Installing git hooks...")
    if not script_path.exists():
        fail(f"{script_path} script does not exist, quitting...")

    git_hooks_path = get_git_hooks_path()

    _install_post_checkout(git_hooks_path)
    _install_pre_checkout(git_hooks_path)

    info("git hooks installed!", delete_last_line=True)


def uninstall(args: argparse.Namespace) -> None:
    info("Uninstalling git hooks...")
    git_hooks_path = get_git_hooks_path()

    _uninstall_post_checkout(git_hooks_path)
    _uninstall_pre_checkout(git_hooks_path)

    info("git hooks uninstalled!", delete_last_line=True)



import argparse

from .migrations import reverse_migrations
from .setup import (
    install,
    uninstall,
)


def main() -> None:
    parser = argparse.ArgumentParser(
        prog="great-scott",
        description="A simple tool to automatically, "
        "when changing git branches, "
        "reverse migrations in a Django application.",
    )
    sp = parser.add_subparsers()

    sp_install = sp.add_parser("install", help="Installs git hooks")
    sp_install.set_defaults(func=install)

    sp_uninstall = sp.add_parser("uninstall", help="Uninstalls git hooks")
    sp_uninstall.set_defaults(func=uninstall)

    sp_reverse = sp.add_parser("reverse", help="Reverses migrations")
    sp_reverse.add_argument("--dst-branch", required=True)
    sp_reverse.set_defaults(func=reverse_migrations)

    try:
        args = parser.parse_args()
        args.func(args)
    except AttributeError:
        parser.print_help()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['great_scott']

package_data = \
{'': ['*']}

entry_points = \
{'console_scripts': ['great-scott = great_scott.__main__:main']}

setup_kwargs = {
    'name': 'great-scott',
    'version': '0.1.0',
    'description': 'Automatically reverses Django migrations when changing git branches',
    'long_description': '<div align="center">\n\n# great-scott\n\n`great-scott` is a tool that can automatically reverse migrations in a Django project, when switching between different GIT branches.\n\n[Installation](#installation) •\n[Usage](#usage) •\n[Removing GIT hooks](#removing-git-hooks)\n\n![doc](img/doc.jpeg)\n\n</div>\n\n\n## Installation\n\n`great-scott` can be installed in 2 easy steps:\n\n1. **Install the tool**\n \n   I recommend installing via [pipx](https://github.com/pypa/pipx):\n   ```sh\n   pipx install great-scott\n   ```\n\n   Alternatively, having the project\'s virtual-env activated, you can of course install the tool directly into virtual-env:\n   ```sh\n   python -m pip install great-scott\n   ```\n\n2. **Setup GIT hooks**\n\n   After successful installation, while in the GIT-managed project folder and with virtual-env activated, execute:\n   ```sh\n   great-scott install\n   ```\n\n\n## Usage\n\nReversing migrations should simply work when the GIT branch is changed:\n\n```sh\n$ git checkout mig-test\nSwitched to branch \'main\'\n👀 Looking for migrations to reverse on mig-test...\n⚠️ reversing migrations for importer (up to 0015)\n⚠️ reversing migrations for permissions (up to 0018)\n⚠️ reversing migrations for subscriptions (up to 0044)\nI have reversed migrations for 3 apps!\n```\n\n👉 If you use GIT integration in any of the popular IDEs, such as PyCharm or VSCode, don\'t forget to properly configure the appropriate python environment there so that automatic reversing of migrations is possible.\n\n\n## Removing GIT hooks\n\nTo remove the GIT hooks, execute the command (while in the project directory and with virtual-env activated):\n```sh\ngreat-scott uninstall\n```\n',
    'author': 'Marcin Górecki',
    'author_email': 'marcin.gorecki@hey.com',
    'maintainer': 'Marcin Górecki',
    'maintainer_email': 'marcin.gorecki@hey.com',
    'url': 'https://github.com/goreckim/great-scott',
    'packages': packages,
    'package_data': package_data,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
